% Author: Yanting Wu

%% Basic Parameters
global T lambda0 lambda1 R1 R2 C alpha lambdaF kappa beta Tnp Vlambda0 Vlambda1 VbSA rho ts NI TMAX armno X X2;

R1=1; R2=2; C=0.5;
beta=0.75;

%% Get lambdas 
% Given some integer k, search for lambda0, lambda1 such that K_opt = k

lambda0_s=0.01:0.05:0.9;  %lambda0 search space
lambda1_s=lambda0_s(1):0.05:0.99; % lambda1 search space

Kopt=[1,2,3,4,8];  % K_opt array, search for lambda0, lambda1 such that the K_opt is some value in the array, infinity is a special one, and shown later
Dopt=ones(6,1);  % flag, when special lambda0, lambda1 for K_opt = k is found, set the flag be zero, such that other matches will be ignored
 
for ii=1:length(lambda0_s)
    for jj=1:length(lambda1_s)
        lambda0=lambda0_s(ii);
        lambda1=lambda1_s(jj);
        alpha=lambda1-lambda0;
        lambdaF=lambda0/(1-alpha);
        kappa=@(n) (1-alpha.^(n+1))*lambdaF;
        
        % Vlambda0, Vlambda1
        nn=1:10000;
        Vlambda0_s=((1-beta.^nn)/(1-beta)*R1...
            +beta.^nn.*(kappa(nn)./(1-beta*lambda1)*R2+kappa(nn)*C*(1-beta)/(1-beta*lambda1)-C))...
            ./...
            (1-beta.^(nn+1).*(1-(1-beta)*kappa(nn)/(1-beta*lambda1)));
        Vlambda0=max(max(Vlambda0_s),R1/(1-beta));
        Vlambda1=max((lambda1*R2-(1-lambda1)*C+beta*(1-lambda1)*Vlambda0)/(1-beta*lambda1), R1/(1-beta));
              
        % VbSA
        VbSA=@(p) (p*R2-(1-p)*C)+beta*(p*Vlambda1+(1-p)*Vlambda0);
        
        rho = calculateRho();

        % Tnp
        Tnp=@(n,p) lambdaF*(1-alpha.^n)+alpha.^n*p;
        % K
        K=find(Tnp(0:10000,lambda0)>=rho,1)-1;  % how many time slots staying in conservative state
        
        if isempty(K) 
            if Dopt(1)==1
            %Vbp?
            display(lambda0);
            display(lambda1);
            display(rho);
            KK=20;
            Vbp=0:KK;
            for kk=0:KK
                Vbp(kk+1)=R1*(1-beta.^kk)/(1-beta)+beta.^kk.*VbSA(Tnp(kk,lambda0));
            end
            plot(0:KK,Vbp);
            title('Infinity');
            pause
            Dopt(1)=0;
            end
            continue;
        end

        for tt=2:6
            if Dopt(tt)==1  && K==Kopt(tt-1)
                Dopt(tt)=0;
                display(lambda0);
                display(lambda1);
                display(K);
                display(rho);        
                %Vbp?
                KK=20;
                Vbp=0:KK;
                for kk=0:KK
                    Vbp(kk+1)=R1*(1-beta.^kk)/(1-beta)+beta.^kk.*VbSA(Tnp(kk,lambda0));
                end
                plot(0:KK,Vbp);
                title(num2str(K));
                pause
                continue
            end
        end
    end
end

%% Run the simulations

% set the lambda0, lambda1 to be the value got from the above section
lambda0=0.16; 
lambda1=0.91;

T=1e7; % the number of time slots
armno=30; % the number of arms, the denotion is L in the paper, K = armno is a special arm, which simulates always sending conservatively case
TMAX = 100; % the number of time slots playing the same arm, the denotion is T in the paper
currentarm=1; % the arm index that is being played now
X = zeros(armno, 1); % normalized discounted reward, in the range of [0, 1]
X2 = zeros(armno,1); % normalized discounted reward square, used in UCBTuned algorithm
NI = zeros(armno, 1); % the number of periods each arm is played
N = 0;
ts =0; % the index of time slot starting playing another arm

% get an initial discounted reward for each arm
for K=0:armno-1
    getDiscountedReward(K);
end
N = N + armno;

iter = 10; % the purpose of this variable is to simulate more times slots than T, and avoid memeory overflow
t_sample=floor(10.^linspace(2,log10(T/TMAX*iter-1),200)); % samples
NI_sample=zeros(length(t_sample)+1,length(NI));
NI_ind=1;
is_UCB_tuned = true;

VV=zeros(armno,2);
VV(:,2)=1/4;
for kk=1:iter
    generateStates;
    ts = 0;
    while ts < T-TMAX     
       if N==t_sample(NI_ind)
           %display(N);
           NI_sample(NI_ind,:)=NI.';
           NI_ind=NI_ind+1;
       end
       
       if is_UCB_tuned == true
            VV(:,1)=X2./NI-(X./NI).^2+sqrt(2*log(N)./NI);     
            [Xmax, Xind] = max((X./NI)+sqrt(log(N)./NI.*min(VV,[],2)));
       else
            [Xmax, Xind] = max((X./NI)+sqrt(2*log(N)./NI));
       end
       
       getDiscountedReward(Xind-1);% array index starts from 1, not 0
       N=N+1;
    end
end

[Nmax,Nind]=max(NI);
display(sprintf('%d,%d,%d,%d\n',Nmax,Nind,T,N));
